package com.xy.service;

import cn.easyes.core.biz.EsPageInfo;
import cn.easyes.core.conditions.LambdaEsQueryWrapper;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.xy.annotate.Runners;
import com.xy.annotate.Timer;
import com.xy.dto.*;
import com.xy.entity.MqttCommand;
import com.xy.mapper.MqttCommandMapper;
import com.xy.producer.MqttProducer;
import com.xy.utils.*;
import com.xy.utils.consts.CommConsts;
import com.xy.utils.enums.CmdTypeEnum;
import com.xy.utils.enums.MqttCommandStatusEnum;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static com.xy.utils.Beans.copy;
import static com.xy.utils.EsBeans.toPageBean;

@Slf4j
@Service
@Api(tags = "mqtt接口")
@RequiredArgsConstructor
public class MqttServiceImpl implements MqttService {

    private final MqttProducer mqttProducer;

    private final MqttCommandMapper mqttCommandMapper;

    private final RedisService<String> redisService;

    private String keyStr = "mqtt:cmd:result:%d";

    private long getSn() {
        return redisService.atomicIncrease(CommConsts.CMD_ATOM_SN);
    }

    private long getWkSn() {
        return redisService.atomicIncrease(CommConsts.CMD_ATOM_WK_SN);
    }

    private List<Tuple.Tuple3<Long, Long, Boolean>> sendMqtt(List<MqttDto> mqttDtos) {
        String time = DataTime.getSring();
        List<Tuple.Tuple3<Long, Long, Boolean>> list = new ArrayList<>(mqttDtos.size());
        mqttDtos.forEach(mqttDto -> {
            //给予默认值
            PaterDto pater = mqttDto.getPater()
                    .setSn(getSn())
                    .setTime(time)
                    .setDeviceId(mqttDto.getDeviceId())
                    .setDebug(mqttDto.getDebug());
            //延迟发布处理
            String delayTime = mqttDto.getDelayTime();
            String topic = mqttDto.getDeviceId() + CommConsts.DEVICE_MQTT_TOPIC_SUFFIX;
            if (!StringUtils.isEmpty(delayTime)) {
                //消息消费时间 - 发布时间的差值 = 距今被消费的秒数
                long m = DataTime.diff(DataTime.toLocal(pater.getTime()), DataTime.toLocal(delayTime), "s");
                topic = String.format("$delayed/%d/%s", m, topic);
                mqttDto.setTimeout(mqttDto.getTimeout() + (int) m);
            }
            //消息过期时间
            pater.setTimeout(mqttDto.getTimeout());
            JSONObject value = JSONUtil.parseObj(pater);
            Long wkSn = pater.getWkSn();
            try {
                //指令记录
                saveMqttCommand(mqttDto, value, pater.getAck() ? MqttCommandStatusEnum.COMMAND_SEND : MqttCommandStatusEnum.COMMAND_EXE_SUCCESS, null);
                //发送消息
                mqttProducer.sendToMqtt(value.toString(), topic, mqttDto.getLevel());
                list.add(new Tuple.Tuple3<>(wkSn, pater.getSn(), true));
            } catch (Exception e) {
                //指令记录
                saveMqttCommand(mqttDto, value, MqttCommandStatusEnum.COMMAND_SEND_FAIL, e.getMessage());
                list.add(new Tuple.Tuple3<>(wkSn, pater.getSn(), false));
            }
        });
        return list;
    }

    @Timer
    @Runners
    @PostMapping("sysnAtoms")
    @ApiOperation("同步sn号redis原子变量")
    public void sysnAtoms() {
        LambdaEsQueryWrapper<MqttCommand> wrapper = new LambdaEsQueryWrapper<MqttCommand>()
                .orderByDesc(MqttCommand::getId);
        EsPageInfo<MqttCommand> mqttCommandEsPageInfo = mqttCommandMapper.pageQuery(wrapper, 0, 1);
        List<MqttCommand> list = mqttCommandEsPageInfo.getList();
        if (!Emptys.check(list)) {
            return;
        }
        MqttCommand mqttCommand = list.get(0);
        redisService.setAtomic(CommConsts.CMD_ATOM_SN, mqttCommand.getId() + 1);
        redisService.setAtomic(CommConsts.CMD_ATOM_WK_SN, JSONUtil.parseObj(mqttCommand.getValue()).getLong("wkSn") + 1);
    }

    @Timer
    @Override
    @ApiOperation("指令发送")
    public R<List<Tuple.Tuple3<Long, Long, Boolean>>> senCommand(List<CommandMqtt> commandMqtts) {
        //发送指令
        List<MqttDto> mqttDtos = new ArrayList<>();
        commandMqtts.forEach(commandMqtt -> {
            JSONObject templetObj = commandMqtt.getTemplet();
            String debug = templetObj.getStr(LambdaUtils.getProperty(MqttDto::getDebug));
            String level = templetObj.getStr(LambdaUtils.getProperty(MqttDto::getLevel));
            String delayTime = templetObj.getStr(LambdaUtils.getProperty(MqttDto::getDelayTime));
            String timeout = templetObj.getStr(LambdaUtils.getProperty(MqttDto::getTimeout));
            String wkSn = templetObj.getStr(LambdaUtils.getProperty(PaterDto::getWkSn));
            String actionType = templetObj.getStr(LambdaUtils.getProperty(PaterDto::getActionType));
            String cmdType = templetObj.getStr(LambdaUtils.getProperty(PaterDto::getCmdType));
            String ack = templetObj.getStr(LambdaUtils.getProperty(PaterDto::getAck));
            JSONObject data = templetObj.getJSONObject(LambdaUtils.getProperty(PaterDto::getData));
            PaterDto paterDto = new PaterDto()
                    .setActionType(actionType)
                    .setCmdType(cmdType)
                    .setAck(Emptys.check(ack) && !ack.equals("$ack$") ? Boolean.valueOf(ack) : Emptys.check(ack) && ack.equals("$ack$") ? true : true)
                    .setWkSn(!Emptys.check(wkSn) ? getWkSn() : "$wkSn$".equals(wkSn) ? getWkSn() : Long.valueOf(wkSn))
                    .setData(data);
            MqttDto mqttDto = new MqttDto()
                    .setDeviceId(commandMqtt.getDeviceId())
                    .setDebug(Emptys.check(debug) && !debug.equals("$debug$") ? Boolean.valueOf(debug) : Emptys.check(debug) && debug.equals("$debug$") ? false : false)
                    .setLevel(Emptys.check(level) && !level.equals("$level$") ? Integer.valueOf(level) : Emptys.check(level) && level.equals("$level$") ? 1 : 1)
                    .setDelayTime(Emptys.check(delayTime) && !delayTime.equals("$delayTime$") ? delayTime : Emptys.check(delayTime) && delayTime.equals("$delayTime$") ? null : null)
                    .setTimeout(Emptys.check(timeout) && !timeout.equals("$timeout$") ? Integer.valueOf(timeout) : Emptys.check(timeout) && timeout.equals("$timeout$") ? 60 : 60)
                    .setPater(paterDto);
            mqttDtos.add(mqttDto);
        });
        List<Tuple.Tuple3<Long, Long, Boolean>> list = sendMqtt(mqttDtos);
        //执行业务
        Map<String, List<MqttDto>> map = new HashMap<>();
        mqttDtos.forEach(mqttDto -> {
            String cmdType = mqttDto.getPater().getCmdType();
            if (!map.containsKey(cmdType)) {
                map.put(cmdType, new ArrayList<>());
            }
            map.get(cmdType).add(mqttDto);
        });
        map.forEach((k, v) -> SpringBeanUtils.getBean(CmdTypeEnum.getEnum(k).getCmdServiceClass()).senCommand(v));
        return R.ok(list);
    }

    @ApiOperation("指令Ack")
    @PostMapping("commandAck")
    public R commandAck(@RequestBody @Validated AckMqttDto ackMqttDto) {
        MqttCommand mqttCommand = mqttCommandMapper.selectById(ackMqttDto.getSn());
        if (!Emptys.check(mqttCommand)) {
            return R.fail("数据不存在");
        }
        //写入es
        mqttCommand.setStatus(Integer.valueOf(ackMqttDto.getStatusEnum().getKey()))
                .setBackClientTime(DataTime.toString(ackMqttDto.getBackClientTime()))
                .setBackServerTime(DataTime.getSring());
        mqttCommandMapper.updateById(mqttCommand);
        //写入redis
        String result = ackMqttDto.getResult().toString();
        if (Emptys.check(ackMqttDto.getResult())) {
            String key = String.format(keyStr, ackMqttDto.getSn());
            redisService.set(key, result, 300);
        }
        return R.ok();
    }

    @ApiOperation("指令结果通知")
    @PostMapping("commandResultBack")
    public R commandResultBack(@RequestBody @Validated MqttDto.ResultBack resultBack) {
        return SpringBeanUtils.getBean(CmdTypeEnum.getEnum(resultBack.getCmdType()).getCmdServiceClass()).resultBack(resultBack);
    }

    @PostMapping("snByCmdAndResult")
    @ApiOperation("sn查询指令和结果数据")
    public R<MqttDto.Vo3> snByCmdAndResult(@RequestBody MqttDto.SnByCmdAndResult snByCmdAndResult) {
        MqttCommand mqttCommand = mqttCommandMapper.selectById(snByCmdAndResult.getSn());
        String key = String.format(keyStr, snByCmdAndResult.getSn());
        String result = redisService.get(key);
        MqttDto.Vo3 vo3 = copy(MqttDto.Vo3.class, mqttCommand).setResult(JSONUtil.parseObj(result));
        return R.ok(vo3);
    }

    @PostMapping("page")
    @ApiOperation("分页查询")
    public R<PageBean<MqttDto.Vo2>> page(@RequestBody MqttDto.Page page) {
        LambdaEsQueryWrapper<MqttCommand> lambdaEsQueryWrapper = new EsQueryWrapper().eqWrapper(page, MqttCommand.class)
                .like(MqttCommand::getError)
                .like(MqttCommand::getValue)
                .ge(MqttCommand::getSendTime, page.getBeginSendTime())
                .le(MqttCommand::getSendTime, page.getEndSendTime())
                .ge(MqttCommand::getTimeout, page.getBeginTimeout())
                .le(MqttCommand::getTimeout, page.getEndTimeout())
                .ge(MqttCommand::getDelayTime, page.getBeginDelayTime())
                .le(MqttCommand::getDelayTime, page.getEndDelayTime())
                .build()
                .orderByDesc(MqttCommand::getSendTime);
        PageBean pageBean = page.getPage();
        EsPageInfo<MqttCommand> mqttCommandEsPageInfo = mqttCommandMapper.pageQuery(lambdaEsQueryWrapper, (int) pageBean.getCurrent(), (int) pageBean.getSize());
        return R.ok(toPageBean(MqttDto.Vo2.class, mqttCommandEsPageInfo));
    }


    /**
     * 新增指令记录
     *
     * @param mqttDto 参数对象
     * @param value   参数json对象
     * @param status  状态
     * @param error   错误信息
     */
    private void saveMqttCommand(MqttDto mqttDto, JSONObject value, MqttCommandStatusEnum status, String error) {
        PaterDto pater = mqttDto.getPater();
        MqttCommand mqttCommand = Beans.copy(Beans.copy(MqttCommand.class, mqttDto), pater)
                .setId(pater.getSn())
                .setStatus(Integer.parseInt(status.getKey()))
                .setValue(value.toString())
                .setSendTime(pater.getTime())
                .setNum(1)
                .setError(error);
        mqttCommandMapper.insert(mqttCommand);
    }
}
